#include <stdio.h>
#include <stdlib.h>
#include "dre_app.h"

int main(int argc, char **argv)
{
    dre_return_t result;
    dre_app_handle *h = NULL;
    dre_app_resource_value_t resource;
    char *name = NULL;
    char *cgroup = NULL;
    char *type = NULL;
    char *limitation = NULL;
    char *exec = argv[0];
    char **p = ++argv;

    int i;
    for(i = 0; p[i]; i++) {
        if(!strcmp(argv[i], "-n") && p[i + 1]) {
            name = argv[++i];
        } else if(!strcmp(argv[i], "--cpu") && p[i + 1]) {
           type = "cpu";
           limitation = argv[++i];
        } else if(!strcmp(argv[i], "--core") && p[i + 1]) {
           type = "core";
           limitation = argv[++i];
        } else if(!strcmp(argv[i], "--mem") && p[i + 1]) {
           type = "mem";
           limitation = argv[++i];
        } else if(!strcmp(argv[i], "--memsw") && p[i + 1]) {
           type = "memsw";
           limitation = argv[++i];
        } else if(!strcmp(argv[i], "-g") && p[i + 1]) {
            cgroup = argv[++i];
        } else goto arg;
    }
    if(!name || !type || !limitation) goto arg;

    /* set resource usage */
    if(!strcmp(type, "cpu")) {
        resource.type = DRE_APP_CONTAINER_RESOURCE_CPU;
        resource.udata.cpu.percent = atoi(limitation);
    } else if(!strcmp(type, "core")) {
        resource.type = DRE_APP_CONTAINER_RESOURCE_CPU_CORE;
        resource.udata.cpu.cores = limitation;
    } else if(!strcmp(type, "mem")) {
        resource.type = DRE_APP_CONTAINER_RESOURCE_MEMORY;
        resource.udata.mem.usage_in_bytes = atoi(limitation);
    } else if(!strcmp(type, "memsw")) {
        resource.type = DRE_APP_CONTAINER_RESOURCE_MEMSW;
        resource.udata.mem.memsw_usage_in_bytes = atoi(limitation);
    } else {
        printf("Invalid resource type: %s\n", type);
        result = DRE_ERR_INVALID_ARGUMENT;
        goto error;
    }

    /* open handler */
    if(name) {
        h = dre_app_open(name);
        if(!h) {
            result = DRE_ERR_NULL_POINTER;
            goto error;
        }
    }

    /* set limitation */
    result = dre_app_set_resource(h, cgroup, resource);
    if(result) goto error;
    printf("Set limitation of %s.\n", type);

    /* close handler */
    if(h) {
        result = dre_app_close(h);
        if(result) goto error;
    }

    return 0;

arg:
    printf("usage: %s -n <name> <resource_type> <limitation> [-g <cgroup>]\n", exec);
    return DRE_ERR_INVALID_ARGUMENT;

error:
    if(h) {
        dre_app_close(h);
    }

    printf("Failed to set resource usage (ErrNo=%d).\n", result);
    return result;
}

