/*
 * (C) Copyright 2009-2013 JST - Japan Science and Technology Agency
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/wait.h>
#include "dre_sys.h"

#ifndef DRE_BUFFER_LENGTH
#define DRE_BUFFER_LENGTH 1024
#endif


/*
 * escape meta charactors(local function)
 */
static dre_return_t
escape_meta_char(char *dest, const char *src)
{
    dre_sys_return_t ret = {0};
    char *p1, *p2;
    char tmp1[DRE_BUFFER_LENGTH] = "";
    char tmp2[DRE_BUFFER_LENGTH] = "";
    int c = 0;
    char meta[] = {'\\', '"', '$', '@', '\'', '(', ')', '^', '|', 
                   '[', ']', '{', '}', ';', '*', '?'};

    if(!src || !dest) {
        ret.error = DRE_ERR_NULL_POINTER;
        goto _DRE_EXIT_;
    }

    strcpy(dest, "");
    strcpy(tmp1, src);
    for(c = 0; c < sizeof(meta); c++) {
        char escaped[3];
        p1 = tmp1;
        sprintf(escaped, "\\%c", meta[c]);

        while(1) {
            p2 = strchr(p1, meta[c]);
            if(p2) {
                if(strlen(tmp2) + strlen(p1) - strlen(p2) < DRE_BUFFER_LENGTH) {
                    strncat(tmp2, p1, strlen(p1) - strlen(p2));
                } else {
                    ret.error = DRE_ERR_REQUEST_TOO_LONG;
                    goto _DRE_EXIT_;
                }

                if(strlen(tmp2) + strlen(escaped) < DRE_BUFFER_LENGTH) {
                    strcat(tmp2, escaped);
                } else {
                    ret.error = DRE_ERR_REQUEST_TOO_LONG;
                    goto _DRE_EXIT_;
                }
                p1 += strlen(p1) - strlen(p2) + sizeof(char);
            } else {
                if(strlen(tmp2) + strlen(p1) < DRE_BUFFER_LENGTH) {
                    strcat(tmp2, p1);
                } else {
                    ret.error = DRE_ERR_REQUEST_TOO_LONG;
                    goto _DRE_EXIT_;
                }
                break;
            }
        }
        strcpy(tmp1, tmp2);
        strcpy(tmp2, "");
    }
    strcpy(dest, tmp1);

_DRE_EXIT_:
    return ret.value;
}


/*
 * create container
 */
dre_return_t
dre_sys_create_container(const char *name, 
                         const dre_sys_data *data)
{
    dre_sys_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-sys create -n ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char option[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(strlen(cmd) + strlen(name) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, name);

    switch(data->type) {
    case DRE_SYS_CONTAINER_TYPE_KVM:
        if(data->udata.kvm.memory_size > 0) {
            snprintf(option, DRE_BUFFER_LENGTH, " -m %d", data->udata.kvm.memory_size);
            if(strlen(cmd) + strlen(option) > DRE_BUFFER_LENGTH) {
                ret.error = DRE_ERR_REQUEST_TOO_LONG;
                goto _DRE_EXIT_;
            }
            strcat(cmd, option);
        }
        if(data->udata.kvm.base && strlen(data->udata.kvm.base)) {
            if(strlen(data->udata.kvm.base) > DRE_BUFFER_LENGTH) {
                ret.error = DRE_ERR_REQUEST_TOO_LONG;
                goto _DRE_EXIT_;
            }
            snprintf(option, DRE_BUFFER_LENGTH, " -b %s", data->udata.kvm.base);
            if(strlen(cmd) + strlen(option) > DRE_BUFFER_LENGTH) {
                ret.error = DRE_ERR_REQUEST_TOO_LONG;
                goto _DRE_EXIT_;
            }
            strcat(cmd, option);
        }
        if(data->udata.kvm.rootfs && strlen(data->udata.kvm.rootfs)) {
            if(strlen(data->udata.kvm.rootfs) > DRE_BUFFER_LENGTH) {
                ret.error = DRE_ERR_REQUEST_TOO_LONG;
                goto _DRE_EXIT_;
            }
            snprintf(option, DRE_BUFFER_LENGTH, " -f %s", data->udata.kvm.rootfs);
            if(strlen(cmd) + strlen(option) > DRE_BUFFER_LENGTH) {
                ret.error = DRE_ERR_REQUEST_TOO_LONG;
                goto _DRE_EXIT_;
            }
            strcat(cmd, option);
        }
        break;
    }

    if(strlen(cmd) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.error = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-SYS-LIB: create: Failed to create '%s'. (ErrNo=%d)", name, ret.value);
    } else {
        DLOG_INFO("DRE-SYS-LIB: create: Created system container '%s'.", name);
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * destroy container
 */
dre_return_t
dre_sys_destroy_container(const char *name)
{
    dre_sys_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-sys destroy -n ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(strlen(cmd) + strlen(name) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, name);
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.error = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-SYS-LIB: destroy: Failed to destroy '%s'. (ErrNo=%d)", name, ret.value);
    } else {
        DLOG_INFO("DRE-SYS-LIB: destroy: Destroyed system container '%s'.", name);
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * open handler
 */
dre_sys_handle *
dre_sys_open(const char *name)
{
    dre_err_t ret = 0;
    dre_sys_handle *h = NULL;
    char cmd[DRE_BUFFER_LENGTH] = "";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char buf[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";
    FILE *p = NULL;

    /* search container */
    strcpy(cmd, "sudo dre-sys search -n ");
    if(strlen(cmd) + strlen(name) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, name);
    strcat(cmd, redirect);
    escape_meta_char(escaped_cmd, cmd);

    ret = -WEXITSTATUS(system(escaped_cmd));
    if(!ret) {
        h = (dre_sys_handle *)malloc(sizeof(dre_sys_handle));
        h->name = (char *)malloc(sizeof(char) * strlen(name));
        strcpy(h->name, name);
        h->type = 0;
    } else {
        ret = DRE_ERR_NULL_POINTER;
        goto _DRE_EXIT_;
    }

    /* open handler */
    strcpy(cmd, "sudo dre-sys info -n ");
    if(strlen(cmd) + strlen(name) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, name);
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    p = popen(escaped_cmd, "r");
    if(p) {
        while(fgets(buf, sizeof(buf), p)) {
            if(strstr(buf, "TYPE")) {
                char *tmp;
                tmp = strtok(buf, " ");
                h->type = atoi(strtok(NULL, " "));
            }
            break;
        }
    } else {
        ret = DRE_ERR_NULL_POINTER;
        goto _DRE_EXIT_;
    }

_DRE_EXIT_:
    if(p) {
        pclose(p);
    }

    if(!h || !(h->type)) {
        DLOG_ERR("DRE-SYS-LIB: open: Failed to open handler of '%s'. (ErrNo=%d)", name, ret);
    } else {
        DLOG_INFO("DRE-SYS-LIB: open: Opened handler of '%s'.", name);
    }
    return h;
}


/*
 * start container
 */
dre_return_t
dre_sys_start(const dre_sys_handle *h)
{
    dre_sys_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-sys start -n ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(strlen(cmd) + strlen(h->name) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, h->name);
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.error = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-SYS-LIB: start: Failed to start '%s'. (ErrNo=%d)", h->name, ret.value);
    } else {
        DLOG_INFO("DRE-SYS-LIB: start: Started system container '%s'.", h->name);
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * stop container
 */
dre_return_t
dre_sys_stop(const dre_sys_handle *h)
{
    dre_sys_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-sys stop -n ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(strlen(cmd) + strlen(h->name) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, h->name);
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.error = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-SYS-LIB: stop: Failed to stop '%s'. (ErrNo=%d)", h->name, ret.value);
    } else {
        DLOG_INFO("DRE-SYS-LIB: stop: Stopped system container '%s'.", h->name);
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * close handler
 */
dre_return_t
dre_sys_close(dre_sys_handle *h)
{
    dre_sys_return_t ret = {0};
    char name[DRE_BUFFER_LENGTH] = "";

    strcpy(name, h->name);

    if(h) {
        if(h->name) {
            free(h->name);
            h->name = NULL;
        } else {
            ret.error = DRE_ERR_NULL_POINTER;
            goto _DRE_EXIT_;
        }
        free(h);
        h = NULL;
    } else {
        ret.error = DRE_ERR_NULL_POINTER;
        goto _DRE_EXIT_;
    }

_DRE_EXIT_:
    if(h) {
        free(h);
        h = NULL;
    }

    if(ret.value) {
        DLOG_ERR("DRE-SYS-LIB: close: Failed to close handler of '%s'. (ErrNo=%d)", name, ret.value);
    } else {
        DLOG_INFO("DRE-SYS-LIB: close: Closed handler of '%s'.", name);
    }

    return ret.value;
}


/*
 * get status
 */
dre_return_t
dre_sys_get_status(const dre_sys_handle *h)
{
    dre_sys_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-sys status -n ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";
    int result = 0;

    if(strlen(cmd) + strlen(h->name) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, h->name);
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    result = WEXITSTATUS(system(escaped_cmd));
    if(result >= -DRE_ERR_INVALID_ARGUMENT) {
        ret.error = -result;
    } else {
        ret.status = result;
    }

    if(ret.value < 0) {
        DLOG_ERR("DRE-SYS-LIB: status: Failed to get status of '%s'. (ErrNo=%d)", h->name, ret.value);
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * get pid of container
 */
dre_return_t
dre_sys_get_pid(const dre_sys_handle *h, int *pid)
{
    dre_sys_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-sys getpid -n ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char buf[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";
    FILE *p;

    if(!h || !pid) {
        ret.error = DRE_ERR_NULL_POINTER;
        goto _DRE_EXIT_;
    }

    if(strlen(cmd) + strlen(h->name) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, h->name);
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    p = popen(escaped_cmd, "r");
    if(p) {
        char *pr = fgets(buf, sizeof(buf), p);
        if(!pr) {
            // TODO: error
        }
        ret.error = -WEXITSTATUS(pclose(p));
        if(!ret.error) {
            *pid = atoi(buf);
        }
    } else {
        ret.error = DRE_ERR_NULL_POINTER;
        goto _DRE_EXIT_;
    }

_DRE_EXIT_:
    if(ret.value) {
        DLOG_ERR("DRE-SYS-LIB: getpid: Failed to get PID of '%s'. (ErrNo=%d)", h->name, ret.value);
    } else {
        DLOG_INFO("DRE-SYS-LIB: getpid: PID of '%s' is %d.", h->name, *pid);
    }

    return ret.value;
}


/*
 * create control group
 */
dre_return_t
dre_sys_create_cgroup(const dre_sys_handle *h, const char *cgroup)
{
    dre_sys_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-sys create -g ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char option[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(!cgroup) {
        ret.error = DRE_ERR_NULL_POINTER;
        goto _DRE_EXIT_;
    }

    if(strlen(cmd) + strlen(cgroup) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, cgroup);

    if(h) {
        sprintf(option, " -n %s", h->name);
        if(strlen(cmd) + strlen(option) > DRE_BUFFER_LENGTH) {
            ret.error = DRE_ERR_REQUEST_TOO_LONG;
            goto _DRE_EXIT_;
        }
        strcat(cmd, option);
    }

    if(strlen(cmd) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.status = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-SYS-LIB: create: Failed to create control group '%s'. (ErrNo=%d)", cgroup, ret.value);
    } else {
        DLOG_INFO("DRE-SYS-LIB: create: Created control group '%s'.", cgroup);
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * destroy control group
 */
dre_return_t
dre_sys_destroy_cgroup(const dre_sys_handle *h, const char *cgroup)
{
    dre_sys_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-sys destroy -g ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char option[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(!cgroup) {
        ret.error = DRE_ERR_NULL_POINTER;
        goto _DRE_EXIT_;
    }

    if(strlen(cmd) + strlen(cgroup) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, cgroup);

    if(h) {
        sprintf(option, " -n %s", h->name);
        if(strlen(cmd) + strlen(option) > DRE_BUFFER_LENGTH) {
            ret.error = DRE_ERR_REQUEST_TOO_LONG;
            goto _DRE_EXIT_;
        }
        strcat(cmd, option);
    }

    if(strlen(cmd) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.status = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-SYS-LIB: destroy: Failed to destroy control group '%s'. (ErrNo=%d)", cgroup, ret.value);
    } else {
        DLOG_INFO("DRE-SYS-LIB: destroy: Destroyed control group '%s'.", cgroup);
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * bind pid to control group
 */
dre_return_t
dre_sys_bind_pid(const dre_sys_handle *h, const char *cgroup, const int pid)
{
    dre_sys_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-sys bind -g ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char option[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(!cgroup) {
        ret.error = DRE_ERR_NULL_POINTER;
        goto _DRE_EXIT_;
    }

    if(strlen(cmd) + strlen(cgroup) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, cgroup);

    sprintf(option, " -pid %d", pid);
    if(strlen(cmd) + strlen(option) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, option);

    if(h) {
        sprintf(option, " -n %s", h->name);
        if(strlen(cmd) + strlen(option) > DRE_BUFFER_LENGTH) {
            ret.error = DRE_ERR_REQUEST_TOO_LONG;
            goto _DRE_EXIT_;
        }
        strcat(cmd, option);
    }

    if(strlen(cmd) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.status = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-SYS-LIB: bind: Failed to bind '%d' to '%s'. (ErrNo=%d)", pid, cgroup, ret.value);
    } else {
        DLOG_INFO("DRE-SYS-LIB: bind: Bound '%d' to '%s'.", pid, cgroup);
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * unbind pid from control group
 */
dre_return_t
dre_sys_unbind_pid(const dre_sys_handle *h, const char *cgroup, const int pid)
{
    dre_sys_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-sys unbind -g ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char option[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(!cgroup) {
        ret.error = DRE_ERR_NULL_POINTER;
        goto _DRE_EXIT_;
    }

    if(strlen(cmd) + strlen(cgroup) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, cgroup);

    sprintf(option, " -pid %d", pid);
    if(strlen(cmd) + strlen(option) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, option);

    if(h) {
        sprintf(option, " -n %s", h->name);
        if(strlen(cmd) + strlen(option) > DRE_BUFFER_LENGTH) {
            ret.error = DRE_ERR_REQUEST_TOO_LONG;
            goto _DRE_EXIT_;
        }
        strcat(cmd, option);
    }

    if(strlen(cmd) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.status = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-SYS-LIB: unbind: Failed to unbind '%d' from '%s'. (ErrNo=%d)", pid, cgroup, ret.value);
    } else {
        DLOG_INFO("DRE-SYS-LIB: unbind: Unbound '%d' from '%s'.", pid, cgroup);
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * get container list
 */
dre_return_t
dre_sys_get_container_list(dre_list_t *list)
{
    dre_sys_return_t ret = {0};
    int i = 0;
    char buf[DRE_BUFFER_LENGTH] = "";
    dre_list_t plist = NULL;
    FILE *p;

    p = popen("sudo dre-sys list", "r");
    if(p) {
        while(fgets(buf, sizeof(buf), p)) {
            buf[strlen(buf) - 1] = '\0';
            plist = (dre_list_t)realloc(plist, sizeof(buf) * (i + 1));
            *(plist + i) = (char *)malloc(sizeof(buf));
            strcpy(*(plist + i), buf);
            i++;
        }
        if(plist) {
            free(*(plist + i));
            *(plist + i) = NULL;
        }
        *list = plist;
        pclose(p);
    } else {
        ret.error = DRE_ERR_NULL_POINTER;
    }

    return ret.value;
}


/*
 * release container list
 */
dre_return_t
dre_sys_release_container_list(dre_list_t list)
{
    dre_sys_return_t ret = {0};
    int i = 0;

    if(*list) {
        while(*(list + i)) {
            free(*(list + i));
            *(list + i) = NULL;
            i++;
        }
        free(list);
        list = NULL;
    } else {
        ret.error = DRE_ERR_NULL_POINTER;
    }

    return ret.value;
}


/*
 * get base image list
 */
dre_return_t
dre_sys_get_base_list(dre_list_t *list)
{
    dre_sys_return_t ret = {0};
    int i = 0;
    char buf[DRE_BUFFER_LENGTH] = "";
    dre_list_t plist = NULL;
    FILE *p;

    p = popen("sudo dre-sys list --base", "r");
    if(p) {
        while(fgets(buf, sizeof(buf), p)) {
            buf[strlen(buf) - 1] = '\0';
            plist = (dre_list_t)realloc(plist, sizeof(buf) * (i + 1));
            *(plist + i) = (char *)malloc(sizeof(buf));
            strcpy(*(plist + i), buf);
            i++;
        }
        if(plist) {
            free(*(plist + i));
            *(plist + i) = NULL;
        }
        *list = plist;
        pclose(p);
    } else {
        ret.error = DRE_ERR_NULL_POINTER;
    }

    return ret.value;
}


/*
 * release base image list
 */
dre_return_t
dre_sys_release_base_list(dre_list_t list)
{
    return dre_sys_release_container_list(list);
}


/*
 * remove base image
 */
dre_return_t
dre_sys_remove_base(const char *base)
{
    dre_sys_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-sys remove --base ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char option[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    sprintf(option, "%s", base);
    if(strlen(cmd) + strlen(option) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, option);
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.error = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-SYS-LIB: remove: Failed to remove base image '%s'. (ErrNo=%d)", base, ret.value);
    } else {
        DLOG_INFO("DRE-SYS-LIB: remove: Removed base image '%s'.", base);
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * start migration target
 */
dre_return_t
dre_sys_immigration_start(const char *name, const int port)
{
    dre_sys_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-target -n ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char option[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(strlen(cmd) + strlen(name) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, name);

    sprintf(option, " -p %d", port);
    if(strlen(cmd) + strlen(option) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, option);
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.error = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-SYS-LIB: immigrate: Failed to immigrate '%s'. (ErrNo=%d)", name, ret.value);
    } else {
        DLOG_INFO("DRE-SYS-LIB: immigrate: Immigrated '%s'.", name);
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * start migration
 */
dre_return_t
dre_sys_migration_start(const dre_sys_handle *h, 
                                  const char *target, const int port)
{
    dre_sys_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-sys migrate -n ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char option[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(strlen(cmd) + strlen(h->name) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, h->name);

    sprintf(option, " -d %s", target);
    if(strlen(cmd) + strlen(option) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, option);

    sprintf(option, " -p %d", port);
    if(strlen(cmd) + strlen(option) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, option);
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.error = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-SYS-LIB: migrate: Failed to migrate '%s' to '%s'. (ErrNo=%d)", h->name, target, ret.value);
    } else {
        DLOG_INFO("DRE-SYS-LIB: migrate: Migrated '%s' to '%s'.", h->name, target);
    }

_DRE_EXIT_:
    return 0;
}


/*
 * get resource usage
 */
dre_return_t
dre_sys_get_resource(const dre_sys_handle *h, const int pid, const char *app,
                               dre_sys_resource_value_t *resource)
{
    dre_sys_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char option[DRE_BUFFER_LENGTH] = "";
    char buf[DRE_BUFFER_LENGTH] = "";
    FILE *p;

    switch(resource->type) {
    case DRE_SYS_CONTAINER_RESOURCE_CPU:
        sprintf(cmd, "sudo dre-sys usage --cpu");
        break;
    case DRE_SYS_CONTAINER_RESOURCE_CPU_CORE:
        sprintf(cmd, "sudo dre-sys usage --core");
        break;
    case DRE_SYS_CONTAINER_RESOURCE_MEMORY:
        sprintf(cmd, "sudo dre-sys usage --mem");
        break;
    case DRE_SYS_CONTAINER_RESOURCE_MEMSW:
        sprintf(cmd, "sudo dre-sys usage --memsw");
        break;
    }

    if(h) {
        sprintf(option, " -n %s", h->name);
        if(strlen(cmd) + strlen(option) > DRE_BUFFER_LENGTH) {
            ret.error = DRE_ERR_REQUEST_TOO_LONG;
            goto _DRE_EXIT_;
        }
        strcat(cmd, option);
    } else {
        ret.error = DRE_ERR_NULL_POINTER;
        goto _DRE_EXIT_;
    }

    if(app) {
        sprintf(option, " -a %s", app);
        if(strlen(cmd) + strlen(option) > DRE_BUFFER_LENGTH) {
            ret.error = DRE_ERR_REQUEST_TOO_LONG;
            goto _DRE_EXIT_;
        }
        strcat(cmd, option);
    }

    if(pid > 0) {
        if(app) {
            sprintf(option, " -vpid %d", pid);
        } else {
            sprintf(option, " -pid %d", pid);
        }
        if(strlen(cmd) + strlen(option) > DRE_BUFFER_LENGTH) {
            ret.error = DRE_ERR_REQUEST_TOO_LONG;
            goto _DRE_EXIT_;
        }
        strcat(cmd, option);
    }

    escape_meta_char(escaped_cmd, cmd);
puts(escaped_cmd);
    p = popen(escaped_cmd, "r");
    if(p) {
        char *pr = fgets(buf, sizeof(buf), p);
        if(!pr) {
            // TODO: error
        }
        ret.error = -WEXITSTATUS(pclose(p));
printf("%d\n", ret.error);
        if(!ret.error) {
            switch(resource->type) {
            case DRE_SYS_CONTAINER_RESOURCE_CPU:
                resource->udata.cpu.percent = atoi(buf);
                break;
            case DRE_SYS_CONTAINER_RESOURCE_CPU_CORE:
                resource->udata.cpu.cores = buf;
                break;
            case DRE_SYS_CONTAINER_RESOURCE_MEMORY:
                resource->udata.mem.usage_in_bytes = atoi(buf);
                break;
            case DRE_SYS_CONTAINER_RESOURCE_MEMSW:
                resource->udata.mem.memsw_usage_in_bytes = atoi(buf);
                break;
            }
        }
    } else {
        ret.error = DRE_ERR_NULL_POINTER;
        goto _DRE_EXIT_;
    }

_DRE_EXIT_:
    if(ret.value) {
        DLOG_ERR("DRE-SYS-LIB: usage: Failed to get resource usage of '%s'. (ErrNo=%d)", h->name, ret.value);
    }

    return (int)ret.value;
}


/*
 * set resource usage
 */
dre_return_t
dre_sys_set_resource(const dre_sys_handle *h, const char *cgroup, const char *app, 
                               const dre_sys_resource_value_t resource)
{
    dre_sys_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char option[DRE_BUFFER_LENGTH] = "";
    char target[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(h) {
        sprintf(option, " -n %s", h->name);
        if(strlen(target) + strlen(option) > DRE_BUFFER_LENGTH) {
            ret.error = DRE_ERR_REQUEST_TOO_LONG;
            goto _DRE_EXIT_;
        }
        strcat(target, option);
    } else {
        ret.error = DRE_ERR_NULL_POINTER;
        goto _DRE_EXIT_;
    }

    if(app) {
        sprintf(option, " -a %s", app);
        if(strlen(target) + strlen(option) > DRE_BUFFER_LENGTH) {
            ret.error = DRE_ERR_REQUEST_TOO_LONG;
            goto _DRE_EXIT_;
        }
        strcat(target, option);
    }

    if(cgroup) {
        sprintf(option, " -g %s", cgroup);
        if(strlen(target) + strlen(option) > DRE_BUFFER_LENGTH) {
            ret.error = DRE_ERR_REQUEST_TOO_LONG;
            goto _DRE_EXIT_;
        }
        strcat(target, option);
    }

    switch(resource.type) {
    case DRE_SYS_CONTAINER_RESOURCE_CPU:
        sprintf(cmd, "sudo dre-sys limit --cpu %d", resource.udata.cpu.percent);
        break;
    case DRE_SYS_CONTAINER_RESOURCE_CPU_CORE:
        sprintf(cmd, "sudo dre-sys limit --core %s", resource.udata.cpu.cores);
        break;
    case DRE_SYS_CONTAINER_RESOURCE_MEMORY:
        sprintf(cmd, "sudo dre-sys limit --mem %d", resource.udata.mem.usage_in_bytes);
        break;
    case DRE_SYS_CONTAINER_RESOURCE_MEMSW:
        sprintf(cmd, "sudo dre-sys limit --memsw %d", resource.udata.mem.memsw_usage_in_bytes);
        break;
    }

    if(strlen(cmd) + strlen(target) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }

    strcat(cmd, target);
    strcat(cmd, redirect);
    escape_meta_char(escaped_cmd, cmd);

    ret.error = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value != 0) {
        goto _DRE_EXIT_;
    }

    DLOG_INFO("DRE-SYS-LIB: limit: Set limitation of resource usage '%s'.", h->name);

_DRE_EXIT_:
    DLOG_ERR("DRE-SYS-LIB: limit: Failed to set limitation of resource usage '%s'. (ErrNo=%d)", cgroup, ret.value);

    return ret.value;
}


/*
 * unset resource usage
 */
dre_return_t
dre_sys_unset_resource(const dre_sys_handle *h, const char *cgroup, const char *app, 
                                 const dre_sys_resource_value_t resource)
{
    dre_sys_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char option[DRE_BUFFER_LENGTH] = "";
    char target[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(h) {
        sprintf(option, " -n %s", h->name);
        if(strlen(target) + strlen(option) > DRE_BUFFER_LENGTH) {
            ret.error = DRE_ERR_REQUEST_TOO_LONG;
            goto _DRE_EXIT_;
        }
        strcat(target, option);
    } else {
        ret.error = DRE_ERR_NULL_POINTER;
        goto _DRE_EXIT_;
    }

    if(app) {
        sprintf(option, " -a %s", app);
        if(strlen(target) + strlen(option) > DRE_BUFFER_LENGTH) {
            ret.error = DRE_ERR_REQUEST_TOO_LONG;
            goto _DRE_EXIT_;
        }
        strcat(target, option);
    }

    if(cgroup) {
        sprintf(option, " -c %s", cgroup);
        if(strlen(target) + strlen(option) > DRE_BUFFER_LENGTH) {
            ret.error = DRE_ERR_REQUEST_TOO_LONG;
            goto _DRE_EXIT_;
        }
        strcat(target, option);
    }

    switch(resource.type) {
    case DRE_SYS_CONTAINER_RESOURCE_CPU:
        sprintf(cmd, "sudo dre-sys unlimit --cpu");
        break;
    case DRE_SYS_CONTAINER_RESOURCE_CPU_CORE:
        sprintf(cmd, "sudo dre-sys unlimit --core");
        break;
    case DRE_SYS_CONTAINER_RESOURCE_MEMORY:
        sprintf(cmd, "sudo dre-sys unlimit --mem");
        break;
    case DRE_SYS_CONTAINER_RESOURCE_MEMSW:
        sprintf(cmd, "sudo dre-sys unlimit --memsw");
        break;
    }

    if(strlen(cmd) + strlen(target) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }

    strcat(cmd, target);
    strcat(cmd, redirect);
    escape_meta_char(escaped_cmd, cmd);

    ret.error = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value != 0) {
        goto _DRE_EXIT_;
    }

    DLOG_INFO("DRE-SYS-LIB: unlimit: Unset limitation of resource usage '%s'.", h->name);

_DRE_EXIT_:
    DLOG_ERR("DRE-SYS-LIB: unlimit: Failed to unset limitation of resource usage '%s'. (ErrNo=%d)", cgroup, ret.value);

    return ret.value;
}


/*
 * suspend container
 */
dre_return_t
dre_sys_suspend(const dre_sys_handle *h)
{
    dre_sys_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-sys suspend -n ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(strlen(cmd) + strlen(h->name) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, h->name);
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.error = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-SYS-LIB: suspend: Failed to suspend '%s'. (ErrNo=%d)", h->name, ret.value);
    } else {
        DLOG_INFO("DRE-SYS-LIB: suspend: Suspended '%s'.", h->name);
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * resume container
 */
dre_return_t
dre_sys_resume(const dre_sys_handle *h)
{
    dre_sys_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-sys resume -n ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(strlen(cmd) + strlen(h->name) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, h->name);
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.error = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-SYS-LIB: resume: Failed to resume '%s'. (ErrNo=%d)", h->name, ret.value);
    } else {
        DLOG_INFO("DRE-SYS-LIB: resume: Resumed '%s'.", h->name);
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * get checkpoint list
 */
dre_return_t
dre_sys_get_checkpoint_list(const dre_sys_handle *h, 
                                      dre_list_t *list)
{
    dre_sys_return_t ret = {0};
    int i = 0;
    char cmd[DRE_BUFFER_LENGTH] = "dre-sys list --checkpoint -n ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char buf[DRE_BUFFER_LENGTH] = "";
    dre_list_t plist = NULL;
    FILE *p;

    if(strlen(cmd) + strlen(h->name) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, h->name);

    escape_meta_char(escaped_cmd, cmd);
    p = popen(escaped_cmd, "r");
    if(p) {
        if(h->type == DRE_SYS_CONTAINER_TYPE_KVM) {
            while(fgets(buf, sizeof(buf), p)) {
                if(!strstr(buf, "ID")) {
                    continue;
                } else {
                    break;
                }
            }
        }
        while(fgets(buf, sizeof(buf), p)) {
            char *tmp, *tag;
            buf[strlen(buf) - 1] = '\0';

            if(h->type == DRE_SYS_CONTAINER_TYPE_KVM) {
                tmp = strtok(buf, " ");
                tag = strtok(NULL, " ");
            } else {
                tag = buf;
            }

            plist = (dre_list_t)realloc(plist, sizeof(buf) * (i + 1));
            *(plist + i) = (char *)malloc(sizeof(buf));
            strcpy(*(plist + i), tag);
            i++;
        }
        if(plist) {
            *(plist + i) = NULL;
        }
        *list = plist;
        pclose(p);
    } else {
        ret.error = DRE_ERR_NULL_POINTER;
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * release checkpoint list
 */
dre_return_t
dre_sys_release_checkpoint_list(dre_list_t list)
{
    return dre_sys_release_container_list(list);
}


/*
 * create checkpoint
 */
dre_return_t
dre_sys_checkpoint(const dre_sys_handle *h, 
                             const char *tag)
{
    dre_sys_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "dre-sys checkpoint -n ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char option[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(strlen(cmd) + strlen(h->name) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, h->name);

    sprintf(option, " -t %s", tag);
    if(strlen(cmd) + strlen(option) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, option);
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.error = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-SYS-LIB: checkpoint: Failed to save checkpoint of '%s' as '%s'. (ErrNo=%d)", h->name, tag, ret.value);
    } else {
        DLOG_INFO("DRE-SYS-LIB: checkpoint: Saved checkpoint of '%s' as '%s'.", h->name, tag);
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * restart from checkpoint
 */
dre_return_t
dre_sys_restart(const dre_sys_handle *h, 
                          const char *tag)
{
    dre_sys_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "dre-sys restart -n ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char option[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(strlen(cmd) + strlen(h->name) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, h->name);

    sprintf(option, " -t %s", tag);
    if(strlen(cmd) + strlen(option) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, option);
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.error = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-SYS-LIB: restart: Failed to restart '%s' from '%s'. (ErrNo=%d)", h->name, tag, ret.value);
    } else {
        DLOG_INFO("DRE-SYS-LIB: restart: Restarted '%s' from '%s'.", h->name, tag);
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * remove checkpoint
 */
dre_return_t
dre_sys_remove_checkpoint(const dre_sys_handle *h, 
                                  const char *tag)
{
    dre_sys_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-sys remove --checkpoint -n ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char option[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(strlen(cmd) + strlen(h->name) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, h->name);

    sprintf(option, " -t %s", tag);
    if(strlen(cmd) + strlen(option) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, option);
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.error = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-SYS-LIB: remove: Failed to remove checkpoint '%s' of '%s'. (ErrNo=%d)", tag, h->name, ret.value);
    } else {
        DLOG_INFO("DRE-SYS-LIB: remove: Removed checkpoint '%s' of '%s'.", tag, h->name);
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * share the container
 */
dre_return_t
dre_sys_share(const dre_sys_handle *h)
{
    dre_sys_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "dre-sys share -n ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char option[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(strlen(cmd) + strlen(h->name) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, h->name);
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.error = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-SYS-LIB: share: Failed to share '%s'. (ErrNo=%d)", h->name, ret.value);
    } else {
        DLOG_INFO("DRE-SYS-LIB: share: Shared '%s'.", h->name);
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * unshare the container
 */
dre_return_t
dre_sys_unshare(const dre_sys_handle *h)
{
    dre_sys_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "dre-sys unshare -n ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char option[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(strlen(cmd) + strlen(h->name) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, h->name);
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.error = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-SYS-LIB: unshare: Failed to unshare '%s'. (ErrNo=%d)", h->name, ret.value);
    } else {
        DLOG_INFO("DRE-SYS-LIB: unshare: Unshared '%s'.", h->name);
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * get snapshot list
 */
dre_return_t
dre_sys_get_snapshot_list(const dre_sys_handle *h, 
                                    dre_list_t *list)
{
    dre_sys_return_t ret = {0};
    int i = 0;
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-sys list --snapshot -n ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char buf[DRE_BUFFER_LENGTH] = "";
    dre_list_t plist = NULL;
    FILE *p;

    if(strlen(cmd) + strlen(h->name) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, h->name);

    escape_meta_char(escaped_cmd, cmd);
    p = popen(escaped_cmd, "r");
    if(p) {
        if(h->type == DRE_SYS_CONTAINER_TYPE_KVM) {
            while(fgets(buf, sizeof(buf), p)) {
                if(!strstr(buf, "ID")) {
                    continue;
                } else {
                    break;
                }
            }
        }
        while(fgets(buf, sizeof(buf), p)) {
            char *tmp, *tag;
            buf[strlen(buf) - 1] = '\0';

            if(h->type == DRE_SYS_CONTAINER_TYPE_KVM) {
                tmp = strtok(buf, " ");
                tag = strtok(NULL, " ");
            } else {
                tag = buf;
            }

            plist = (dre_list_t)realloc(plist, sizeof(buf) * (i + 1));
            *(plist + i) = (char *)malloc(sizeof(buf));
            strcpy(*(plist + i), tag);
            i++;
        }

        if(plist) {
            *(plist + i) = NULL;
        }

        *list = plist;
        pclose(p);
    } else {
        ret.error = DRE_ERR_NULL_POINTER;
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * release snapshot list
 */
dre_return_t
dre_sys_release_snapshot_list(dre_list_t list)
{
    return dre_sys_release_container_list(list);
}


/*
 * save snapshot
 */
dre_return_t
dre_sys_save_snapshot(const dre_sys_handle *h, 
                                const char *tag)
{
    dre_sys_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-sys save -n ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char option[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(strlen(cmd) + strlen(h->name) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, h->name);

    sprintf(option, " -t %s", tag);
    if(strlen(cmd) + strlen(option) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, option);
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.error = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-SYS-LIB: save: Failed to save snapshot of '%s' as '%s'. (ErrNo=%d)", h->name, tag, ret.value);
    } else {
        DLOG_INFO("DRE-SYS-LIB: save: Saved snapshot of '%s' as '%s'.", h->name, tag);
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * load snapshot
 */
dre_return_t
dre_sys_load_snapshot(const dre_sys_handle *h, 
                                const char *tag)
{
    dre_sys_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-sys load -n ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char option[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(strlen(cmd) + strlen(h->name) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, h->name);

    sprintf(option, " -t %s", tag);
    if(strlen(cmd) + strlen(option) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, option);
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.error = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-SYS-LIB: load: Failed to load snapshot '%s' of '%s'. (ErrNo=%d)", tag, h->name, ret.value);
    } else {
        DLOG_INFO("DRE-SYS-LIB: load: Loaded snapshot '%s' of '%s'.", tag, h->name);
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * remove snapshot
 */
dre_return_t
dre_sys_remove_snapshot(const dre_sys_handle *h, 
                                  const char *tag)
{
    dre_sys_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-sys remove --snapshot -n ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char option[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(strlen(cmd) + strlen(h->name) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, h->name);

    sprintf(option, " -t %s", tag);
    if(strlen(cmd) + strlen(option) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, option);
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.error = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-SYS-LIB: remove: Failed to remove snapshot '%s' of '%s'. (ErrNo=%d)", tag, h->name, ret.value);
    } else {
        DLOG_INFO("DRE-SYS-LIB: remove: Removed snapshot '%s' of '%s'.", tag, h->name);
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * commit snapshot
 */
dre_return_t
dre_sys_commit_snapshot(const dre_sys_handle *h, 
                                  const char *tag, 
                                  const char *base)
{
    dre_sys_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-sys commit -n ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char option[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(strlen(cmd) + strlen(h->name) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, h->name);

    sprintf(option, " -t %s", tag);
    if(strlen(cmd) + strlen(option) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, option);

    sprintf(option, " -b %s", base);
    if(strlen(cmd) + strlen(option) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, option);
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.error = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-SYS-LIB: commit: Failed to commit snapshot '%s' to '%s'. (ErrNo=%d)", tag, h->name, ret.value);
    } else {
        DLOG_INFO("DRE-SYS-LIB: commit: Committed snapshot '%s' to '%s'.", tag, h->name);
    }

_DRE_EXIT_:
    return ret.value;
}

