/*
 * (C) Copyright 2009-2013 JST - Japan Science and Technology Agency
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#ifndef _DRE_APP_H_
#define _DRE_APP_H_

#include "dre_error.h"
#include "daware.h"

// list
#ifndef DRE_LIST
#define DRE_LIST
typedef char * dre_name_t;
typedef dre_name_t * dre_list_t;
#endif

#ifndef DRE_RETURN
#define DRE_RETURN
typedef int dre_return_t;
#endif


// container type
typedef enum {
    DRE_APP_CONTAINER_TYPE_LXC = 1
} dre_app_type_t;

// status type
typedef enum {
    DRE_APP_CONTAINER_STATUS_STOP = 11,
    DRE_APP_CONTAINER_STATUS_RUNNING,
    DRE_APP_CONTAINER_STATUS_SUSPENDED,
    DRE_APP_CONTAINER_STATUS_BUSY
} dre_app_status_t;

// return type
typedef union {
    dre_return_t value;
    dre_app_status_t status;
    dre_err_t error;
} dre_app_return_t;

// resource type
typedef enum {
    DRE_APP_CONTAINER_RESOURCE_CPU = 1,
    DRE_APP_CONTAINER_RESOURCE_CPU_CORE,
    DRE_APP_CONTAINER_RESOURCE_MEMORY,
    DRE_APP_CONTAINER_RESOURCE_MEMSW
} dre_app_resource_t;

// container handler
typedef struct {
    dre_app_type_t type;
    char *name;
} dre_app_handle;

// container data
typedef struct {
    dre_app_type_t type;
    union {
        struct {
            int port;
            const char *base;
        } lxc;
    } udata;
} dre_app_data;

// resource usage
typedef struct {
    dre_app_resource_t type;
    union {
        union {
            int percent;
            char *cores;
        } cpu;
        union {
            int usage_in_bytes;
            int memsw_usage_in_bytes;
        } mem;
    } udata;
} dre_app_resource_value_t;

// process id
typedef struct {
    int pid;
    int vpid;
} dre_app_pid_t;

// pid list
typedef dre_app_pid_t ** dre_app_pid_list_t;


#ifdef __cplusplus
extern "C" {
#endif

extern dre_return_t dre_app_create_container(const char *name, const dre_app_data *data);
extern dre_return_t dre_app_destroy_container(const char *name);

extern dre_app_handle *dre_app_open(const char *name);
extern dre_return_t dre_app_start(const dre_app_handle *h);
extern dre_return_t dre_app_stop(const dre_app_handle *h);
extern dre_return_t dre_app_close(dre_app_handle *h);
extern dre_return_t dre_app_get_status(const dre_app_handle *h);
extern dre_return_t dre_app_get_pid(const dre_app_handle *h, int *pid);

extern dre_return_t dre_app_create_cgroup(const dre_app_handle *h, const char *cgroup);
extern dre_return_t dre_app_destroy_cgroup(const dre_app_handle *h, const char *cgroup);

extern dre_return_t dre_app_bind_pid(const dre_app_handle *h, const char *cgroup, const int pid);
extern dre_return_t dre_app_bind_vpid(const dre_app_handle *h, const char *cgroup, const int vpid);
extern dre_return_t dre_app_unbind_pid(const dre_app_handle *h, const char *cgroup, const int pid);
extern dre_return_t dre_app_unbind_vpid(const dre_app_handle *h, const char *cgroup, const int vpid);

extern dre_return_t dre_app_register_pid(const dre_app_handle *h, const int pid);
extern dre_return_t dre_app_register_vpid(const dre_app_handle *h, const int vpid);
extern dre_return_t dre_app_unregister_pid(const dre_app_handle *h, const int pid);
extern dre_return_t dre_app_unregister_vpid(const dre_app_handle *h, const int vpid);

extern dre_return_t dre_app_get_registered_pid_list(const dre_app_handle *h, dre_app_pid_list_t *list);
extern dre_return_t dre_app_release_registered_pid_list(dre_app_pid_list_t list);

extern dre_return_t dre_app_get_container_list(dre_list_t *list);
extern dre_return_t dre_app_release_container_list(dre_list_t list);

extern dre_return_t dre_app_get_base_list(dre_list_t *list);
extern dre_return_t dre_app_release_base_list(dre_list_t list);
extern dre_return_t dre_app_remove_base(const char *base);

extern dre_return_t dre_app_get_resource(const dre_app_handle *h, const int pid, const char *cgroup, dre_app_resource_value_t *resource);
extern dre_return_t dre_app_set_resource(const dre_app_handle *h, const char *cgroup, const dre_app_resource_value_t resource);
extern dre_return_t dre_app_unset_resource(const dre_app_handle *h, const char *cgroup, const dre_app_resource_value_t resource);

extern dre_return_t dre_app_suspend(const dre_app_handle *h);
extern dre_return_t dre_app_resume(const dre_app_handle *h);

extern dre_return_t dre_app_get_checkpoint_list(const dre_app_handle *h, dre_list_t *list);
extern dre_return_t dre_app_release_checkpoint_list(dre_list_t list);
extern dre_return_t dre_app_checkpoint(const dre_app_handle *h, const char *tag);
extern dre_return_t dre_app_restart(const dre_app_handle *h, const char *tag);
extern dre_return_t dre_app_remove_checkpoint(const dre_app_handle *h, const char *tag);

extern dre_return_t dre_app_get_snapshot_list(const dre_app_handle *h, dre_list_t *list);
extern dre_return_t dre_app_release_snapshot_list(dre_list_t list);
extern dre_return_t dre_app_save_snapshot(const dre_app_handle *h, const char *tag);
extern dre_return_t dre_app_load_snapshot(const dre_app_handle *h, const char *tag);
extern dre_return_t dre_app_remove_snapshot(const dre_app_handle *h, const char *tag);
extern dre_return_t dre_app_commit_snapshot(const dre_app_handle *h, const char *tag, const char *base);

#ifdef __cplusplus
}
#endif

#endif // _DRE_APP_H_

