#include <stdio.h>
#include <stdlib.h>
#include "dre_app.h"

int main(int argc, char **argv)
{
    dre_app_resource_value_t resource;
    dre_return_t result;
    dre_app_handle *h = NULL;
    char *name = NULL;
    char *type = NULL;
    char *cgroup = NULL;
    char *exec = argv[0];
    int pid = -1;
    char **p = ++argv;

    int i;
    for(i = 0; p[i]; i++) {
        if(!strcmp(argv[i], "-n") && p[i + 1]) {
            name = argv[++i];
        } else if(!strcmp(argv[i], "--cpu")) {
            type = "cpu";
        } else if(!strcmp(argv[i], "--core")) {
            type = "core";
        } else if(!strcmp(argv[i], "--mem")) {
            type = "mem";
        } else if(!strcmp(argv[i], "--memsw")) {
            type = "memsw";
        } else if(!strcmp(argv[i], "-g") && p[i + 1]) {
            cgroup = argv[++i];
        } else if(!strcmp(argv[i], "--pid") && p[i + 1]) {
            pid = atoi(argv[++i]);
        } else goto arg;
    }
    if(!name || !type) goto arg;

    /* open handler */
    h = dre_app_open(name);
    if(!h) {
        result = DRE_ERR_NULL_POINTER;
        goto error;
    }

    /* register process */
    if(pid > 0) {
        result = dre_app_register_pid(h, pid);
        if(result) goto error;
    }

    /* get resource usage */
    if(!strcmp(type, "cpu")) {
        resource.type = DRE_APP_CONTAINER_RESOURCE_CPU;
        result = dre_app_get_resource(h, pid, cgroup, &resource);
        if(result) goto error;
        printf("cpu usage:    %d%%\n", resource.udata.cpu.percent);
    } else if(!strcmp(type, "core")) {
        resource.type = DRE_APP_CONTAINER_RESOURCE_CPU_CORE;
        result = dre_app_get_resource(h, pid, cgroup, &resource);
        if(result) goto error;
        printf("assined cpu core(s):    %s\n", resource.udata.cpu.cores);
    } else if(!strcmp(type, "mem")) {
        resource.type = DRE_APP_CONTAINER_RESOURCE_MEMORY;
        result = dre_app_get_resource(h, pid, cgroup, &resource);
        if(result) goto error;
        printf("memory usage: %dBytes\n", resource.udata.mem.usage_in_bytes);
    } else if(!strcmp(type, "memsw")) {
        resource.type = DRE_APP_CONTAINER_RESOURCE_MEMORY;
        result = dre_app_get_resource(h, pid, cgroup, &resource);
        if(result) goto error;
        printf("memory+swap usage: %dBytes\n", resource.udata.mem.memsw_usage_in_bytes);
    } else {
        printf("Invalid resource type: %s\n", type);
        result = DRE_ERR_INVALID_ARGUMENT;
        goto error;
    }

    /* unregister process */
    if(pid > 0) {
        result = dre_app_unregister_pid(h, pid);
        if(result) goto error;
    }

    /* close handler */
    result = dre_app_close(h);
    if(result) goto error;

    return 0;

arg:
    printf("usage: %s -n <name> <resource_type> [--pid <pid>] [-g <cgroup>]\n", exec);
    return DRE_ERR_INVALID_ARGUMENT;

error:
    if(h) {
        dre_app_close(h);
    }

    printf("Failed to get resource usage (ErrNo=%d).\n", result);
    return result;
}

