#include <stdio.h>
#include <stdlib.h>
#include "dre_sys.h"

int main(int argc, char **argv)
{
    dre_sys_resource_value_t resource;
    dre_return_t result;
    dre_sys_handle *h = NULL;
    char *name = NULL;
    char *type = NULL;
    char *cgroup = NULL;
    char *app = NULL;
    char *exec = argv[0];
    int pid = -1;
    char **p = ++argv;

    int i;
    for(i = 0; p[i]; i++) {
        if(!strcmp(argv[i], "-n") && p[i + 1]) {
            name = argv[++i];
        } else if(!strcmp(argv[i], "--cpu")) {
            type = "cpu";
        } else if(!strcmp(argv[i], "--core")) {
            type = "core";
        } else if(!strcmp(argv[i], "--mem")) {
            type = "mem";
        } else if(!strcmp(argv[i], "--memsw")) {
            type = "memsw";
        } else if(!strcmp(argv[i], "-g") && p[i + 1]) {
            cgroup = argv[++i];
        } else if(!strcmp(argv[i], "-a") && p[i + 1]) {
            app = argv[++i];
        } else if(!strcmp(argv[i], "--pid") && p[i + 1]) {
            pid = atoi(argv[++i]);
        } else goto arg;
    }
    if(!name || !type) goto arg;

    /* open handler */
    h = dre_sys_open(name);
    if(!h) {
        result = DRE_ERR_NULL_POINTER;
        goto error;
    }

    /* get resource usage */
    if(!strcmp(type, "cpu")) {
        resource.type = DRE_SYS_CONTAINER_RESOURCE_CPU;
        result = dre_sys_get_resource(h, pid, app, &resource);
        if(result) goto error;
        printf("cpu usage:    %d%%\n", resource.udata.cpu.percent);
    } else if(!strcmp(type, "core")) {
        resource.type = DRE_SYS_CONTAINER_RESOURCE_CPU_CORE;
        result = dre_sys_get_resource(h, pid, app, &resource);
        if(result) goto error;
        printf("assined cpu core(s):    %s\n", resource.udata.cpu.cores);
    } else if(!strcmp(type, "mem")) {
        resource.type = DRE_SYS_CONTAINER_RESOURCE_MEMORY;
        result = dre_sys_get_resource(h, pid, app, &resource);
        if(result) goto error;
        printf("memory usage: %dBytes\n", resource.udata.mem.usage_in_bytes);
    } else if(!strcmp(type, "memsw")) {
        resource.type = DRE_SYS_CONTAINER_RESOURCE_MEMORY;
        result = dre_sys_get_resource(h, pid, app, &resource);
        if(result) goto error;
        printf("memory+swap usage: %dBytes\n", resource.udata.mem.memsw_usage_in_bytes);
    } else {
        printf("Invalid resource type: %s\n", type, app);
        result = DRE_ERR_INVALID_ARGUMENT;
        goto error;
    }

    /* close handler */
    result = dre_sys_close(h);
    if(result) goto error;

    return 0;

arg:
    printf("usage: %s -n <name> <resource_type> [--pid <pid>] [-g <cgroup>] [-a <application_container>]\n", exec);
    return DRE_ERR_INVALID_ARGUMENT;

error:
    if(h) {
        dre_sys_close(h);
    }

    printf("Failed to get resource usage (ErrNo=%d).\n", result);
    return result;
}

